from dataclasses import dataclass
from enum import Enum

import google.protobuf.message
from nibiru_proto.proto.dex.v1 import tx_pb2 as dex_events  # noqa
from nibiru_proto.proto.lockup.v1 import tx_pb2 as lockup_events  # noqa
from nibiru_proto.proto.perp.v1 import event_pb2 as perp_events
from nibiru_proto.proto.pricefeed import tx_pb2 as pricefeed_events
from nibiru_proto.proto.stablecoin import events_pb2 as stablecoin_events  # noqa
from nibiru_proto.proto.vpool.v1 import event_pb2 as vpool_events


class EventType(Enum):
    """
    The events enum type shows the type of events available to parse from the nibiruwebsocket object.
    """

    # Perp events
    PositionChangedEvent = perp_events.PositionChangedEvent
    PositionSettledEvent = perp_events.PositionSettledEvent
    PositionLiquidatedEvent = perp_events.PositionLiquidatedEvent
    FundingRateChangedEvent = perp_events.FundingRateChangedEvent

    # Vpool events
    ReserveSnapshotSavedEvent = vpool_events.ReserveSnapshotSavedEvent
    SwapQuoteForBaseEvent = vpool_events.SwapQuoteForBaseEvent
    SwapBaseForQuoteEvent = vpool_events.SwapBaseForQuoteEvent
    MarkPriceChanged = vpool_events.MarkPriceChanged

    # # Dex event
    # PoolJoinedEvent = dex_events.EventPoolJoined
    # PoolCreatedEvent = dex_events.EventPoolCreated
    # PoolExitedEvent = dex_events.EventPoolExited
    # AssetsSwappedEvent = dex_events.EventAssetsSwapped

    # Lockup event
    # LockEvent = lockup_events.EventLock
    # UnlockInitiatedEvent = lockup_events.EventUnlockInitiated
    # UnlockEvent = lockup_events.EventUnlock

    # Pricefeed events
    OracleUpdatePriceEvent = pricefeed_events.EventOracleUpdatePrice
    PairPriceUpdatedEvent = pricefeed_events.EventPairPriceUpdated

    # Stablecoin events
    # TransferEvent = stablecoin_events.EventTransfer
    # MintStableEvent = stablecoin_events.EventMintStable
    # BurnStableEvent = stablecoin_events.EventBurnStable
    # MintNIBIEvent = stablecoin_events.EventMintNIBI
    # BurnNIBIEvent = stablecoin_events.EventBurnNIBI
    # RecollateralizeEvent = stablecoin_events.EventRecollateralize
    # BuybackEvent = stablecoin_events.EventBuyback

    # Bank
    Transfer = "transfer"

    def get_full_path(self):
        if isinstance(self.value, str):
            return self.value
        proto_message: google.protobuf.message.Message = self.value
        return str(proto_message.DESCRIPTOR.full_name)


@dataclass
class EventCaptured:
    event_type: EventType
    payload: dict



from grpc import Channel
from nibiru_proto.proto.vpool.v1 import query_pb2 as vpool_type
from nibiru_proto.proto.vpool.v1 import query_pb2_grpc as vpool_query
from nibiru_proto.proto.vpool.v1.vpool_pb2 import Direction as pbDirection

from nibiru.common import Direction
from nibiru.query_clients.util import QueryClient


class VpoolQueryClient(QueryClient):
    """
    VPool allows to query the endpoints made available by the Nibiru Chain's VPOOL module.
    """

    def __init__(self, channel: Channel):
        self.api = vpool_query.QueryStub(channel)

    def reserve_assets(self, pair: str):
        req = vpool_type.QueryReserveAssetsRequest(pair=pair)
        return self.query(self.api.ReserveAssets, req)

    def all_pools(self):
        req = vpool_type.QueryAllPoolsRequest()
        resp = self.query(self.api.AllPools, req)
        for _, prices in enumerate(resp["prices"]):
            prices["index_price"] = cast_str_to_float_safely(prices["index_price"])
            prices["twap_mark"] = cast_str_to_float_safely(prices["twap_mark"])
        return resp

    def base_asset_price(self, pair: str, direction: Direction, base_asset_amount: str):
        dir_pb = pbDirection.DIRECTION_UNSPECIFIED
        if direction == Direction.ADD:
            dir_pb = pbDirection.ADD_TO_POOL
        elif direction == Direction.REMOVE:
            dir_pb = pbDirection.REMOVE_FROM_POOL

        req = vpool_type.QueryBaseAssetPriceRequest(
            pair=pair, direction=dir_pb, base_asset_amount=base_asset_amount
        )
        return self.query(self.api.BaseAssetPrice, req)


def cast_str_to_float_safely(number_str: str) -> float:
    try:
        number = float(number_str)
    except:
        number = float(0)
    return number



from grpc import Channel
from nibiru_proto.proto.pricefeed import query_pb2 as pf_type
from nibiru_proto.proto.pricefeed import query_pb2_grpc as pf_query

from nibiru.query_clients.util import QueryClient


class PricefeedQueryClient(QueryClient):
    """
    Pricefeed allows to query the endpoints made available by the Nibiru Chain's Pricefeed module.
    """

    def __init__(self, channel: Channel):
        self.api = pf_query.QueryStub(channel)

    def params(self):
        req = pf_type.QueryParamsRequest()
        return self.query(self.api.QueryParams, req)

    def price(self, pair_id: str):
        req = pf_type.QueryPriceRequest(
            pair_id=pair_id,
        )
        return self.query(self.api.QueryPrice, req)

    def prices(self):
        req = pf_type.QueryPricesRequest()
        return self.query(self.api.QueryPrices, req)

    def raw_prices(self, pair_id: str):
        req = pf_type.QueryRawPricesRequest(
            pair_id=pair_id,
        )
        return self.query(self.api.QueryRawPrices, req)

    def oracles(self, pair_id: str):
        req = pf_type.QueryOraclesRequest(
            pair_id=pair_id,
        )
        return self.query(self.api.QueryOracles, req)

    def markets(self):
        req = pf_type.QueryMarketsRequest()
        return self.query(self.api.QueryMarkets, req)



from google.protobuf.json_format import MessageToDict
from grpc import Channel
from nibiru_proto.proto.cosmos.auth.v1beta1 import query_pb2 as auth_type
from nibiru_proto.proto.cosmos.auth.v1beta1 import query_pb2_grpc as auth_query

from nibiru.query_clients.util import QueryClient


class AuthQueryClient(QueryClient):
    """
    AuthQueryClient allows to query the endpoints made available by the Nibiru Chain's auth module.
    """

    def __init__(self, channel: Channel):
        self.api = auth_query.QueryStub(channel)

    def account(self, address: str) -> dict:
        proto_output = self.query(
            api_callable=self.api.Account,
            req=auth_type.QueryAccountRequest(address=address),
            should_deserialize=False,
        )

        output = MessageToDict(proto_output)
        return output

    def accounts(self) -> dict:
        proto_output = self.query(
            api_callable=self.api.Accounts,
            req=auth_type.QueryAccountRequest(),
            should_deserialize=False,
        )

        output = MessageToDict(proto_output)
        return output



import base64
import json
import threading
import time
from multiprocessing import Queue
from typing import List

import google.protobuf.json_format
import google.protobuf.message
from websocket import WebSocketApp

import nibiru.query_clients.util
from nibiru import Network
from nibiru.event_specs import EventCaptured, EventType
from nibiru.utils import clean_nested_dict, init_logger

ERROR_TIMEOUT_SLEEP = 3


class NibiruWebsocket:
    queue: Queue = None
    captured_events_type: List[List[str]]

    def __init__(
        self,
        network: Network,
        captured_events_type: List[EventType] = [],
    ):
        """
        The nibiru listener provides an interface to easily connect and handle subscription to the events of a nibiru
        chain.
        """

        self.websocket_url = network.websocket_endpoint
        if self.websocket_url is None:
            raise ValueError(
                "No websocket endpoint provided. Construct the network object setting up the "
                "`websocket_endpoint` endpoint"
            )

        # We use a dictionary for faster search
        self.captured_events_type: dict[str, EventType] = {
            captured_event.get_full_path(): captured_event
            for captured_event in captured_events_type
        }
        self.queue = Queue()
        self.logger = init_logger("ws-logger")

    def start(self):
        """
        Start the websocket and fill the queue with events.
        """

        self._ws = WebSocketApp(
            self.websocket_url,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
        )
        threading.Thread(
            target=self._ws.run_forever,
            daemon=True,
            name=f"Nibiru websocket @ {self.websocket_url}",
        ).start()

    def _on_open(self, _: WebSocketApp):
        self.logger.info("WebSocket starting")
        self._subscribe()

    def _on_error(self, ws: WebSocketApp, error: Exception):
        self.logger.error(f"Closing websocket, error {error}")
        self.logger.exception(error)
        ws.close()
        time.sleep(ERROR_TIMEOUT_SLEEP)
        ws.run_forever()

    def _on_message(self, _: WebSocketApp, message: str):
        """
        Parse the message and filter through them using the captured event type.
        Put these filtered event in the queue.

        Args:
            _ (WebSocketApp): No idea what this is
            message (str): The message in a utf-8 data received from the server
        """
        log = json.loads(message).get("result", {})
        if not log:
            return

        if log["query"] == "tm.event='Tx'":
            self._handle_txs(log)

        if log["query"] == "tm.event='NewBlock'":
            self._handle_block(log)

    def _handle_block(self, log: dict) -> None:
        """
        Read the log for the begin and end block and put relevant events in the queue.

        Args:
            log (dict): The logs
        """

        block_begin_events = log["data"]["value"]["result_begin_block"]["events"]
        block_end_events = log["data"]["value"]["result_end_block"]["events"]

        block_height = log["data"]["value"]["block"]["header"]["height"]

        for event in block_begin_events + block_end_events:
            self._handle_event(
                block_height=block_height, tx_hash=None, event=event, base64_decode=True
            )

    def _handle_txs(self, log: dict):
        """
        Read the log for the transactions in a block and put relevant events in the queue.

        Args:
            log (dict): The logs
        """
        events = log.get("events")
        if events is None:
            return

        block_height = int(log["data"]["value"]["TxResult"]["height"])
        tx_hash = events["tx.hash"][0]

        events = json.loads(log["data"]["value"]["TxResult"]["result"]["log"])[0]

        for event in events["events"]:
            self._handle_event(block_height, tx_hash, event)

    def _handle_event(
        self, block_height: int, tx_hash: str, event: dict, base64_decode: bool = False
    ):
        """
        Read an event and put it in the queue if it's part of the captured event type.

        Args:
            block_height (int): The height of the block
            tx_hash (str): The hash of the transaction if it's coming from a transaction
            event (dict): The event to parse
            base64_decode (bool, optional): Whether to base64 decode the attributes and key of the event.
                Defaults to False.
        """
        # This below is faster since self.captured_events_type.keys() are hashed

        if base64_decode:

            def decode(value):
                return base64.b64decode(value).decode('utf-8')

        else:

            def decode(value):
                return value

        if event["type"] in self.captured_events_type:

            event_payload = {
                decode(attribute["key"]): decode(attribute["value"]).strip('"')
                for attribute in event["attributes"]
            }

            if not isinstance(
                self.captured_events_type[event["type"]].value,
                str,
            ):
                proto_message = google.protobuf.json_format.Parse(
                    json.dumps(clean_nested_dict(event_payload)),
                    self.captured_events_type[event["type"]].value(),
                    ignore_unknown_fields=True,
                )

                event_payload = nibiru.query_clients.util.deserialize(
                    proto_message,
                    no_sdk_transformation=True,
                )

            event_payload["block_height"] = block_height
            event_payload["tx_hash"] = tx_hash

            self.queue.put(EventCaptured(event["type"], event_payload))

    def _subscribe(self):

        tm_events = {
            "new_block_value": "NewBlock",
            "tx_value": "Tx",
        }

        for _, event_name in tm_events.items():

            self._ws.send(
                json.dumps(
                    {
                        "jsonrpc": "2.0",
                        "method": "subscribe",
                        "id": 1,
                        "params": {"query": f"tm.event='{event_name}'"},
                    }
                )
            )



from nibiru.query_clients.auth import AuthQueryClient  # noqa
from nibiru.query_clients.dex import DexQueryClient  # noqa
from nibiru.query_clients.epoch import EpochQueryClient  # noqa
from nibiru.query_clients.perp import PerpQueryClient  # noqa
from nibiru.query_clients.pricefeed import PricefeedQueryClient  # noqa
from nibiru.query_clients.vpool import VpoolQueryClient  # noqa



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nibiru', 'nibiru.msg', 'nibiru.query_clients']

package_data = \
{'': ['*']}

install_requires = \
['aiocron>=1.8,<2.0',
 'aiohttp>=3.8.1,<4.0.0',
 'asyncio>=3.4.3,<4.0.0',
 'bech32>=1.2.0,<2.0.0',
 'bip32>=3.1,<4.0',
 'coincurve>=17.0.0,<18.0.0',
 'ecdsa>=0.18.0,<0.19.0',
 'eip712-structs>=1.1.0,<2.0.0',
 'grpcio>=1.47.0,<2.0.0',
 'hdwallets>=0.1.2,<0.2.0',
 'mnemonic>=0.20,<0.21',
 'nibiru-proto>=0.14.1,<0.15.0',
 'pre-commit>=2.20.0,<3.0.0',
 'protobuf==3.20.1',
 'pytest>=7.1.2,<8.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'requests>=2.28.1,<3.0.0',
 'shutup>=0.2.0,<0.3.0',
 'websocket-client>=1.4.0,<2.0.0']

setup_kwargs = {
    'name': 'nibiru',
    'version': '0.4.3',
    'description': 'Python SDK for interacting with Nibiru.',
    'long_description': '# Nibiru Python SDK - `nibiru`                           <!-- omit in toc -->\n\n> Python-based client for interacting with the Nibiru blockchain.\n\n<!-- Badges -->\n\n[![Nibiru Test workflow][tests-badge]][tests-workflow]\n[![PyPI Version][pypi-image]][pypi-url]\n[![][documentation-image]][documentation-url]\n[![][discord-badge]][discord-url]\n[![][stars-image]][stars-url]\n[![MIT license][license-badge]][license-link]\n\n<!-- Badges links -->\n\n<!-- pypi -->\n[pypi-image]: https://img.shields.io/pypi/v/nibiru\n[pypi-url]: https://pypi.org/project/nibiru/\n[stars-image]: https://img.shields.io/github/stars/NibiruChain?style=social\n[stars-url]: https://github.com/NibiruChain\n[documentation-image]: https://readthedocs.org/projects/nibiru-py/badge/?version=latest\n[documentation-url]: https://nibiru-py.readthedocs.io/en/latest/?badge=latest\n[discord-badge]: https://img.shields.io/badge/Nibiru%20Chain-%237289DA.svg?style=&logo=discord&logoColor=white\n[discord-url]: https://discord.gg/sgPw8ZYfpQ\n[license-badge]: https://img.shields.io/badge/License-MIT-blue.svg\n[license-link]: https://github.com/NibiruChain/py-sdk/blob/master/LICENSE\n[tests-badge]: https://github.com/NibiruChain/py-sdk/actions/workflows/pytests.yml/badge.svg\n[tests-workflow]: https://github.com/NibiruChain/py-sdk/actions/workflows/pytests.yml\n\nThe `nibiru` package allows you to index, query, and send transactions on the Nibiru Blockchain using Python. It provides access to market data for analysis, visualization, indicator development, algorithmic trading, strategy backtesting, bot programming, and related software engineering.\n\nThe package is intended to be used by coders, developers, technically-skilled traders and  data-scientists for building trading algorithms.\n\n#### README Contents\n\n- [User Guidelines](#user-guidelines)\n  - [Documentation Website](#documentation-website)\n  - [Installation from `PyPI`](#installation-from-pypi)\n- [Development Guidelines](#development-guidelines)\n  - [Setting up a professional dev environment with `pyenv` and `poetry`](#setting-up-a-professional-dev-environment-with-pyenv-and-poetry)\n    - [Pyenv for managing multiple Python interpreters](#pyenv-for-managing-multiple-python-interpreters)\n  - [Installing `poetry` for dependency resolution and publishing packages](#installing-poetry-for-dependency-resolution-and-publishing-packages)\n  - [Installing external dependencies](#installing-external-dependencies)\n  - [Running tests](#running-tests)\n      - [Setting environment variables](#setting-environment-variables)\n      - [Running the tests with `poetry` + `pytest`](#running-the-tests-with-poetry--pytest)\n      - [(option B). Install the `nibiru` package with `pip`](#option-b-install-the-nibiru-package-with-pip)\n  - [Makefile and Protocol Buffers](#makefile-and-protocol-buffers)\n    - [Other dependencies](#other-dependencies)\n    - [Generating types wth protobuf](#generating-types-wth-protobuf)\n  - [Linting](#linting)\n  - [Gotchas](#gotchas)\n\n# User Guidelines\n\n## Documentation Website\n\nDocumentation can be found here: [Nibiru-py documentation](https://nibiru-py.readthedocs.io/en/latest/index.html)\n\n- Learn more about opening and managing your spot and perp positions [here](https://nibiru-py.readthedocs.io/en/latest/nibiru.sdks.tx.html#nibiru-sdks-tx-package)\n- Learn about querying the chain using the Sdk [here](https://nibiru-py.readthedocs.io/en/latest/nibiru.clients.html#nibiru-clients-package)\n\n## Installation from `PyPI`\n\n```sh\npip install nibiru  # requires Python 3.9\n```\n\nYou may need to update `pip` to get this to run:\n\n```sh\npython -m pip install --upgrade pip\n```\n\n<!-- NOTE --------- Deperecating this section as the examples don\'t work.\n\n## Usage Instructions\n\nThe [examples directory](https://github.com/NibiruChain/py-sdk/tree/master/examples) contains runnable examples that showcase how to use the package.\n- Requires Python 3.9+\n- Requires a running instance of the Nibiru blockchain\n\n```bash\n$ pipenv shell\n$ pipenv install\n```\n\n```sh\n# connecting to Nibiru Exchange API and create a new pool\n$ python examples/chain_client/dex/create_pool.py\n\n# sending a msg with bank transfer signs and posts a transaction to the Nibiru Chain\n$ python examples/chain_client/msg_send.py\n```\n\nUpgrade `pip` to the latest version, if you see these warnings:\n  ```\n  WARNING: Value for scheme.platlib does not match. Please report this to <https://github.com/pypa/pip/issues/10151>\n  WARNING: Additional context:   user = True   home = None   root = None   prefix = None\n  ```\n-->\n\n---\n\n# Development Guidelines\n\nOur recommended setup is `pyenv` in combination with `poetry`.\n\n- `pyenv` is a tool for installing and managing Python interpreters. This will let you seamlessly switch between Python versions.\n- `poetry` is used for managing virtual environments, dependency resolution, package installations, package building, and package publishing.\n- We assume you\'re on a Unix machine such as WSL2 Ubuntu, MacOS, or a common Linux distro.\n\nCurrently, `nibiru` is created with Python 3.9.13. It may be compatible with higher versions, but we only run end-to-end tests in 3.9.13.\n\n## Setting up a professional dev environment with `pyenv` and `poetry`\n\n### Pyenv for managing multiple Python interpreters\n\nIf you\'re on MacOS or a common Linux distro, you can install `pyenv` with brew.\n\n```sh\nbrew install pyenv\n```\n\nYou\'ll then need to add the following snippet to your shell config, e.g. your `.bash_profile`, `.bashrc`, or `.zshrc`.\n\n```sh\nexport PYENV_ROOT="$HOME/.pyenv"\nexport PATH="$PYENV_ROOT/bin:$PATH"\neval "$(pyenv init -)"\neval "$(pyenv init --path)"\n```\n\nAfter using `source` on your config or restarting the shell, you should have the `pyenv` root command.\n\nThe command use to install any version of python is `pyenv install`. Display additional info for this command with `pyenv install --help`.\n\n```sh\npyenv install 3.9.13 # example for nibiru\n```\n\nOnce you have a version installed, you can print out the versions on your machine with:\n\n```sh\npyenv versions\n```\n\n```\n# example output\n  system\n* 3.9.13 (set by /home/realu/.python-version)\n  3.10.4\n```\n\nIn this example, I have 2 different interpreters installed on my machine. The one with the `*` is currently set as my **global interpreter**. This is set manually using the `pyenv global` command.\n\n```sh\npyenv global 3.10.4   # switches the global interpreter to 3.10.4\n```\n\nYou can verify this works as expected using `python --version`. You may be familiar with using `python3` as the command instead of `python`. With `pyenv`, this is not necessary.\n\nAdditional usage and installation instructions can be found in the [pyenv repo](https://github.com/pyenv/pyenv).\n\n## Installing `poetry` for dependency resolution and publishing packages\n\nReference: [Poetry docs](https://python-poetry.org/docs/)\n\nPoetry can be installed with both `curl` and `pip`. We recommended using `curl` so that it will be global to your machine.\n\nNOTE We highly, highly, highly recommend that you DO NOT use `brew` to install `poetry`.\nIf you use `brew`, it\'s going to install directly to your system, which prevents you from being able to leverage `pyenv` to seamlessly switch between Python interpreters.\n\n```sh\n# installation with pip: recommended option in tandem with pyenv\npip install poetry\n```\n\n```sh\n# For UNIX systems - installation with curl\ncurl -sSL https://install.python-poetry.org/ | python -\n```\n\nAfter this installation command, add the `poetry` binary to the path in your shell config (if it\'s not done automatically).\n\n```sh\nexport PATH=$PATH:$HOME/.poetry/bin\n```\n\n## Installing external dependencies\n\nThe `nibiru` project is defined by its `pyproject.toml`. At the root of the repo, simply call:\n\n```sh\npoetry install\n```\n\nThis will resolve dependencies between each of the project\'s packages and install them into a virtual environment.\n\n## Running tests\n\n#### Setting environment variables\n\nThere\'s currently a "devnet" running in GCP that the CI workflows use. You can find these secrets at [this notion page](https://www.notion.so/nibiru/Resources-and-Repo-Configs-b31aa8074a2b419d80b0c946ed5efab0) if you have access to it or contact one of the `CODEOWNERS` (@Unique-Divine, @matthiasmatt, @nibiruheisenberg).\nThis is useful so that you can run every part of the package code without needing to visit other repositories.\n\nSet up a `.env` file to set environment variables for the tests.\nThe variables used in the CI build can be found in the `env` section of the [`pytests.yml` workflow](.github/workflows/pytests.yml):\n\n```yaml\njobs:\n  tests:\n    env:\n      # https://www.notion.so/nibiru/Resources-and-Repo-Configs-b31aa8074a2b419d80b0c946ed5efab0\n      CHAIN_ID: ${{ secrets.CHAIN_ID }}\n      HOST: ${{ secrets.HOST }}\n      VALIDATOR_MNEMONIC: ${{ secrets.VALIDATOR_MNEMONIC }}\n      GRPC_PORT: ${{ secrets.GRPC_PORT }}\n      LCD_PORT: ${{ secrets.LCD_PORT }}\n```\n\nYou\'ll need an `.env` configuration like this.\n\n```sh\n# Example configuration for the Nibiry Python SDK\nHOST="..."\nVALIDATOR_MNEMONIC="..."\nORACLE_MNEMONIC="..."\nGRPC_PORT="..."\nLCD_PORT="..."\nCHAIN_ID="..."\nNETWORK_INSECURE=true\n```\n\n#### Running the tests with `poetry` + `pytest`\n\nAfter following the instructions for setting up `poetry`, you can run the tests with `poetry run pytest`:\n\n```sh\npoetry run pytest -p no:warnings # silences warnings\n```\n\n#### (option B). Install the `nibiru` package with `pip`\n\n  ```sh\n  # from local\n  # build and install\n  pip install .\n\n  # from local build\n  pip uninstall nibiru\n  pip install nibiru --no-index --find-links /path/to/nibiru/py-sdk/dist\n\n  # from pypi\n  pip uninstall nibiru\n  pip install nibiru\n  ```\n\n## Makefile and Protocol Buffers\n\n### Other dependencies\n\nTo run shell scripts and commands in the `Makefile`, you\'ll need to install the following tools depending on your operating system.\n\n- **Ubuntu**\n\n  ```bash\n  sudo apt install python3.X-dev autoconf automake build-essential libffi-dev libtool pkg-config\n  ```\n\n- **macOS**\n\n  ```bash\n  brew install autoconf automake libtool\n  ```\n\n- **Fedora**\n\n  ```bash\n  sudo dnf install python3-devel autoconf automake gcc gcc-c++ libffi-devel libtool make pkgconfig\n  ```\n\n### Generating types wth protobuf\n\nThe objective is to run `make proto-gen`, which simply executes `scripts/protocgen.sh`.\n\nIn order to do this, you\'ll need to install a few packages on your system.\n\n```sh\npython -m pip install --user grpcio-tools\npip install mypy-protobuf\n```\n\nIf you get a permissions error such as\n\n```\nrm: cannot remove \'proto/proto/epochs/query.proto\': Permission denied\n```\n\ncall `sudo chown -R [USER-NAME] proto` using the name of user directory.\nYou can find the value for `[USER-NAME]` quickly by running `whoami`. In other words, this should work:\n\n```sh\nsudo chown -R $(whoami) proto\n```\n\nYou\'re done generating types once you\'ve successfully called\n\n```sh\nmake proto-gen\npoetry build # equivalently, you can run `python -m build`\n```\n\n## Linting\n\nEnable git hook which will perform linting before each commit:\n\n```shell\npoetry run pre-commit install\n```\n\nThis will keep your code clean.\n\n## Gotchas\n\nThe `protobuf` package must be version 3.20.x or lower. Otherwise, the following error appears at runtime.\n\n```\nnibiru/clients/__init__.py:1: in <module>\n    from nibiru.clients.dex import Dex  # noqa\nnibiru/clients/dex.py:8: in <module>\n    from nibiru.proto.dex.v1 import query_pb2 as dex_type\nnibiru/proto/dex/v1/query_pb2.py:16: in <module>\n    from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2\n../../../anaconda3/envs/divine/lib/python3.9/site-packages/google/api/annotations_pb2.py:30: in <module>\n    from google.api import http_pb2 as google_dot_api_dot_http__pb2\n../../../anaconda3/envs/divine/lib/python3.9/site-packages/google/api/http_pb2.py:48: in <module>\n    _descriptor.FieldDescriptor(\n../../../anaconda3/envs/divine/lib/python3.9/site-packages/google/protobuf/descriptor.py:560: in __new__\n    _message.Message._CheckCalledFromGeneratedFile()\nE   TypeError: Descriptors cannot not be created directly.\nE   If this call came from a _pb2.py file, your generated code is out of date and must be regenerated with protoc >= 3.19.0.\nE   If you cannot immediately regenerate your protos, some other possible workarounds are:\nE    1. Downgrade the protobuf package to 3.20.x or lower.\nE    2. Set PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION=python (but this will use pure-Python parsing and will be much slower).\nE\nE   More information: https://developers.google.com/protocol-buffers/docs/news/2022-05-06#python-updates\n```\n\n## Usage instructions for publishing.\n\nYou specify updates to publish using the commit (or PR) title with `bump-[version-keyword]`.\nFor the `poetry version` command, ysing any bump rule with a valid semver string will change the version inside `pyproject.toml`. For example,\n\n```\npoetry version patch # moves from x.y.14 to x.y.15\npoetry version minor # moves from x.5.z to x.6.0\npoetry version major # moves from 3.y.z to 4.0.0\n```\nThe list of bump rules includes:\npatch, minor, major, prepatch, preminor, premajor, prerelease.\n\nSo the list of available keywords you an put in a PR includes\n- `bump-patch`:\n- `bump-patch`: 0.0.0 → 0.0.1\n- `bump-minor`: 0.0.* → 0.1.0\n- `bump-major`: 0.*.* → 1.0.0\n- `bump-prepatch`: 0.0.0 → 0.0.1-alpha0\n- `bump-prerelease`: equivalent to `bump-prepatch`\n- `bump-preminor`: 0.0.* → 0.1.0-alpha0\n- `bump-premajor`: 0.*.* → 1.0.0-alpha0\n\nThese guidelines are in the release.yml for future reference.\n',
    'author': 'Nibiru Chain',
    'author_email': 'dev@nibiru.fi',
    'maintainer': 'NibiruHeisenberg',
    'maintainer_email': 'dev@nibiru.fi',
    'url': 'https://github.com/NibiruChain/sdk-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from google.protobuf.json_format import MessageToDict
from grpc import Channel
from nibiru_proto.proto.epochs import query_pb2 as epoch_type
from nibiru_proto.proto.epochs import query_pb2_grpc as epoch_query

from nibiru.query_clients.util import QueryClient


class EpochQueryClient(QueryClient):
    """
    EpochQueryClient allows to query the endpoints made available by the Nibiru Chain's epoch module.
    This module is used to time out certain events like for example funding rate payments.
    """

    def __init__(self, channel: Channel):
        self.api = epoch_query.QueryStub(channel)

    def current_epoch(self, epoch_identifier: str) -> dict:
        proto_output = self.query(
            api_callable=self.api.CurrentEpoch,
            req=epoch_type.QueryCurrentEpochRequest(identifier=epoch_identifier),
            should_deserialize=False,
        )

        output = MessageToDict(proto_output)
        output["currentEpoch"] = int(output["currentEpoch"])
        return output

    def epoch_infos(self) -> dict:
        proto_output = self.query(
            api_callable=self.api.EpochInfos,
            req=epoch_type.QueryEpochsInfoRequest(),
            should_deserialize=False,
        )
        output = MessageToDict(proto_output)
        return output



from nibiru.msg.bank import MsgDelegate, MsgSend, MsgWithdrawDelegatorReward  # noqa
from nibiru.msg.dex import (  # noqa
    MsgCreatePool,
    MsgExitPool,
    MsgJoinPool,
    MsgSwapAssets,
)
from nibiru.msg.perp import (  # noqa
    MsgAddMargin,
    MsgClosePosition,
    MsgLiquidate,
    MsgOpenPosition,
    MsgRemoveMargin,
)
from nibiru.msg.pricefeed import MsgPostPrice  # noqa



import json
import logging
import sys
from datetime import datetime
from typing import Any, Callable, Dict, List, Union

from google.protobuf.timestamp_pb2 import Timestamp

# number of decimal places
PRECISION = 18


# reimplementation of cosmos-sdk/types/decimal.go
def to_sdk_dec(dec: float) -> str:
    '''
    create a decimal from an input decimal.
    valid must come in the form:
        (-) whole integers (.) decimal integers
    examples of acceptable input include:
        -123.456
        456.7890
        345
        -456789

    NOTE - An error will return if more decimal places
    are provided in the string than the constant Precision.

    CONTRACT - This function does not mutate the input str.
    '''
    dec_str = str(dec)

    if len(dec_str) == 0:
        raise TypeError(f'Expected decimal string but got: {dec_str}')

    # first extract any negative symbol
    neg = False
    if dec_str[0] == '-':
        neg = True
        dec_str = dec_str[1:]

    if len(dec_str) == 0:
        raise TypeError(f'Expected decimal string but got: {dec_str}')

    strs = dec_str.split('.')
    len_decs = 0
    combined_str = strs[0]

    if len(strs) == 2:  # has a decimal place
        len_decs = len(strs[1])
        if len_decs == 0 or len(combined_str) == 0:
            raise TypeError(f'Expected decimal string but got: {dec_str}')
        combined_str += strs[1]
    elif len(strs) > 2:
        raise TypeError(f'Expected decimal string but got: {dec_str}')

    if len_decs > PRECISION:
        raise TypeError(
            f'value \'{dec_str}\' exceeds max precision by {PRECISION-len_decs} decimal places: max precision {PRECISION}'
        )

    # add some extra zero's to correct to the Precision factor
    zeros_to_add = PRECISION - len_decs
    zeros = '0' * zeros_to_add
    combined_str += zeros

    try:
        int(combined_str, 10)
    except ValueError as err:
        raise ValueError(
            f'failed to set decimal string with base 10: {combined_str}'
        ) from err

    if neg:
        return '-' + combined_str

    return combined_str


def from_sdk_dec_24(dec_str: str) -> float:
    return float(dec_str) * 1e-24


def from_sdk_dec_n(dec_str: str, n: int = 6) -> float:
    return float(dec_str) * 10 ** (-n)


def format_fields_nested(
    object: Union[list, dict], fn: Callable[[Any], Any], fields: list[str]
) -> Union[list, dict]:
    """
    Format the fields inside a nested dictionary with the function provided

    Args:
        object (Union[list, dict]): The object to format
        fn (Callable[[Any], Any]): The function to format objects with
        fields (list[str]): The fields to format

    Returns:
        Union[list, dict]: The output formatted
    """
    if type(object) == dict:
        output = {}
        for k, v in object.items():
            if type(v) in (dict, list):
                output[k] = format_fields_nested(v, fn, fields)
            else:
                if k in fields:
                    output[k] = fn(v)
                else:
                    output[k] = v

        return output

    if type(object) == list:
        output = []

        for element in object:
            if type(object) in (dict, list):
                output.append(format_fields_nested(element, fn, fields))
            else:
                output.append(element)

        return output


def from_sdk_dec(dec_str: str) -> float:
    if dec_str is None or dec_str == '':
        return 0

    if '.' in dec_str:
        raise TypeError(f'expected a decimal string but got {dec_str} containing \'.\'')

    try:
        int(dec_str)
    except ValueError as err:
        raise ValueError(f'failed to convert {dec_str} to a number') from err

    neg = False
    if dec_str[0] == '-':
        neg = True
        dec_str = dec_str[1:]

    input_size = len(dec_str)
    bz_str = ''
    # case 1, purely decimal
    if input_size <= PRECISION:
        # 0. prefix
        bz_str = '0.'

        # set relevant digits to 0
        bz_str += '0' * (PRECISION - input_size)

        # set final digits
        bz_str += dec_str
    else:
        # inputSize + 1 to account for the decimal point that is being added
        dec_point_place = input_size - PRECISION

        bz_str = dec_str[:dec_point_place]  # pre-decimal digits
        bz_str += '.'  # decimal point
        bz_str += dec_str[dec_point_place:]  # pre-decimal digits

    if neg:
        bz_str = '-' + bz_str

    return float(bz_str)


def to_sdk_int(i: float) -> str:
    return str(int(i))


def from_sdk_int(int_str: str) -> int:
    return int(int_str)


def toPbTimestamp(dt: datetime):
    ts = Timestamp()
    ts.FromDatetime(dt)
    return ts


class ColoredFormatter(logging.Formatter):

    fmt = "%(asctime)s|%(levelname)s|%(funcName)s| %(message)s"

    white = "\x1b[97;20m"
    grey = "\x1b[38;20m"
    green = "\x1b[32;20m"
    cyan = "\x1b[36;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    FORMATS = {
        logging.DEBUG: fmt.format(green, reset),
        logging.INFO: fmt.format(cyan, reset),
        logging.WARNING: fmt.format(yellow, reset),
        logging.ERROR: fmt.format(red, reset),
        logging.CRITICAL: fmt.format(bold_red, reset),
    }

    def format(self, record: logging.LogRecord):
        """Formats a record for the logging handler.

        Args:
            record (logging.LogRecord): Represents an instance of an event being
                logged.
        """
        log_format = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_format, datefmt="%H:%M:%S")
        return formatter.format(record=record)


def init_logger(name: str) -> logging.Logger:
    """
    Simple logger to use throughout the test suite.

    Examples:
    ```python
    from nibiru.utils import init_logger
    LOGGER = init_logger("test-logger")
    LOGGER.info("successfully executed tx staking command")
    LOGGER.debug("debugging error message")
    ```

    Log levels include: [debug, info, warning, error, critical]

    Args:
        name (str): Name of the logger

    Returns:
        logging.Logger: The logger object
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # Logs to stdout so we can at least see logs in GHA.
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)

    handler.setFormatter(fmt=ColoredFormatter())
    logger.addHandler(handler)
    return logger


def clean_nested_dict(dictionary: Union[List, Dict, str]) -> Dict:
    """
    Takes a nested dictionnary with some values being string json values and convert it into a proper nested
    dictionary.

    Eg ::

        {
            "transaction_fee": "{\"denom\":\"unusd\",\"amount\":\"0\"}",
            "funding_payment": "0.000000000000000000",
            "realized_pnl": "0.000000000000000000",
            "bad_debt": "{\"denom\":\"unusd\",\"amount\":\"0\"}",
            "trader_address": "nibi1zaavvzxez0elundtn32qnk9lkm8kmcsz44g7xl",
            "margin": "{\"denom\":\"unusd\",\"amount\":\"10\"}",
            "exchanged_position_size": "0.004999999999999500",
            "tx_hash": "12E496C996E39820B0807857CB7C19674BDD12DC8D789647D68C50BBB8C7D9CF"
        }

    becomes ::

        {
            "transaction_fee": {
                "denom": "unusd",
                "amount": "0"
            },
            "funding_payment": "0.000000000000000000",
            "realized_pnl": "0.000000000000000000",
            "bad_debt": {
                "denom": "unusd",
                "amount": "0"
            },
            "trader_address": "nibi1zaavvzxez0elundtn32qnk9lkm8kmcsz44g7xl",
            "margin": {
                "denom": "unusd",
                "amount": "10"
            },
            "exchanged_position_size": "0.004999999999999500",
            "tx_hash": "12E496C996E39820B0807857CB7C19674BDD12DC8D789647D68C50BBB8C7D9CF"
        }

    Args:
        dictionary (Union[List, Dict, str]): The dictionary to be converted.

    Returns:
        Dict: A converted dictionary.
    """

    if isinstance(dictionary, str):
        dictionary = json.loads(dictionary)

    for key, value in dictionary.items():
        if isinstance(value, str):
            if value[0] == "{" and value[-1] == "}":
                dictionary[key] = clean_nested_dict(value)
            else:
                dictionary[key] = value

    return dictionary



from typing import Dict, List, Union

import google.protobuf.message
from google.protobuf import message as protobuf_message
from google.protobuf.json_format import MessageToDict
from grpc import UnaryUnaryMultiCallable
from grpc._channel import _InactiveRpcError

from nibiru.exceptions import QueryError
from nibiru.utils import from_sdk_dec, from_sdk_int

PROTOBUF_MSG_BASE_ATTRS: List[str] = (
    dir(protobuf_message.Message)
    + ['Extensions', 'FindInitializationErrors', '_CheckCalledFromGeneratedFile']
    + ['_extensions_by_name', '_extensions_by_number']
)
"""PROTOBUF_MSG_BASE_ATTRS (List[str]): The default attributes and methods of
an instance of the 'protobuf.message.Message' class.
"""


def camel_to_snake(camel: str):
    return ''.join(
        ['_' + char.lower() if char.isupper() else char for char in camel]
    ).lstrip('_')


def dict_keys_from_camel_to_snake(d):
    """
    Transform all keys from the dictionnary from camelcase to snake case.

    Args:
        d (dict): The dictionary to transform

    Returns:
        dict: The dictionary transformed
    """
    if isinstance(d, list):
        return [
            dict_keys_from_camel_to_snake(i) if isinstance(i, (dict, list)) else i
            for i in d
        ]
    return {
        camel_to_snake(a): dict_keys_from_camel_to_snake(b)
        if isinstance(b, (dict, list))
        else b
        for a, b in d.items()
    }


def deserialize(
    pb_msg: protobuf_message.Message, no_sdk_transformation: bool = False
) -> dict:
    """Deserializes a proto message into a dictionary.

    - sdk.Dec values are converted to floats.
    - sdk.Int values are converted to ints.
    - Missing fields become blank strings.

    Args:
        pb_msg (protobuf.message.Message)
        no_sdk_transformation (bool): Wether to bypass the sdk transformation. Default to False

    Returns:
        dict: 'pb_msg' as a JSON-able dictionary.
    """
    if not isinstance(pb_msg, protobuf_message.Message):
        raise TypeError(f"expted protobuf Message for 'pb_msg', not {type(pb_msg)}")
    custom_dtypes: Dict[str, bytes] = {
        str(field[1]): field[0].GetOptions().__getstate__().get("serialized", None)
        for field in pb_msg.ListFields()
    }
    serialized_output = {}
    expected_fields: List[str] = [
        attr for attr in dir(pb_msg) if attr not in PROTOBUF_MSG_BASE_ATTRS
    ]

    for _, attr in enumerate(expected_fields):

        attr_search = pb_msg.__getattribute__(attr)
        custom_dtype = custom_dtypes.get(str(attr_search))

        if custom_dtype is not None:

            if "sdk/types.Dec" in str(custom_dtype):
                if no_sdk_transformation:
                    serialized_output[str(attr)] = float(pb_msg.__getattribute__(attr))
                else:
                    serialized_output[str(attr)] = from_sdk_dec(
                        pb_msg.__getattribute__(attr)
                    )
            elif "sdk/types.Int" in str(custom_dtype):
                if no_sdk_transformation:
                    serialized_output[str(attr)] = int(pb_msg.__getattribute__(attr))
                else:
                    serialized_output[str(attr)] = from_sdk_int(
                        pb_msg.__getattribute__(attr)
                    )
            elif "Int" in str(custom_dtype):  # Used for sdk.Coin message normalization
                serialized_output[str(attr)] = int(pb_msg.__getattribute__(attr))
            else:
                try:
                    val = pb_msg.__getattribute__(attr)
                    if hasattr(val, '__len__') and not isinstance(val, str):
                        updated_vals = []
                        for v in val:
                            updated_vals.append(deserialize(v))
                        serialized_output[str(attr)] = updated_vals
                    else:
                        serialized_output[str(attr)] = deserialize(val)
                except:
                    serialized_output[str(attr)] = pb_msg.__getattribute__(attr)
        elif (custom_dtype is None) and (attr_search == ''):
            serialized_output[str(attr)] = ""
        else:
            serialized_output[str(attr)] = deserialize(pb_msg.__getattribute__(attr))

    return serialized_output


def deserialize_exp(proto_message: protobuf_message.Message) -> dict:
    """
    Take a proto message and convert it into a dictionnary.
    sdk.Dec values are converted to be consistent with txs.

    Args:
        proto_message (protobuf.message.Message)

    Returns:
        dict
    """
    output = MessageToDict(proto_message)

    is_sdk_dec = {
        field.camelcase_name: "types.Dec" in str(field.GetOptions())
        for field in proto_message.DESCRIPTOR.fields
    }

    for field in proto_message.DESCRIPTOR.fields:
        if field.message_type is not None:
            # This is another proto object
            try:
                output[field.camelcase_name] = deserialize_exp(
                    proto_message.__getattribute__(field.camelcase_name)
                )
            except AttributeError:
                output[field.camelcase_name] = output[field.camelcase_name]

        elif is_sdk_dec[field.camelcase_name]:
            output[field.camelcase_name] = from_sdk_dec(output[field.camelcase_name])

    return dict_keys_from_camel_to_snake(output)


class QueryClient:
    def query(
        self,
        api_callable: UnaryUnaryMultiCallable,
        req: google.protobuf.message.Message,
        should_deserialize: bool = True,
    ) -> Union[dict, google.protobuf.message.Message]:

        try:
            output: google.protobuf.message.Message = api_callable(req)
            if should_deserialize:
                return deserialize(output)
            return output
        except _InactiveRpcError as err:
            raise QueryError(
                f"Error on {str(api_callable._method).split('/')[-1][:-1]}: {err._state.details}"
            ) from None
